#include <tango.h>
#include "MappedAttr.h"
#include "MapAttributeAllocator.h"

#define ELEMLEN 1024

using namespace std;
using namespace Tango;
			 
namespace TangoDevMapper_ns
{

MapAttributeAllocator::MapAttributeAllocator(string entry, string defaultDevname)
{
	//char failure[265];
	attrs.clear();
	err = "No error";
	string devname;
	string attname;
	string dev_and_att, attlist;
	vector<string> devparts;
	vector<string> attatt;
	vector<string> elems;
	bool error = false;
	unsigned int i;
	MappedAttr *attr = NULL;
	
	attatt = split(entry, ':');
	if(attatt.size() != 2)
	{
		_failure_reason = "there is no ':' separator between the device name and the\n"
				"  list of attributes or there are more than one ':' in the string \n  \""
				+ entry + "\"";
		_invalid = error = true;
	}
	else /* two parts correctly separated by a ':' */
	{
		dev_and_att = attatt[0];
		attlist = attatt[1];
		
		devparts = split(dev_and_att, '/');
		if(devparts.size() == 1) /* only the attribute name */
		{
			devname = defaultDevname;
			attname = devparts[0];
		}
		else if(devparts.size() == 4) /* device name specified, e.g. test/device/1 */
		{
			devname = devparts[0] + "/" + devparts[1] + "/" + devparts[2];
			attname = devparts[3];
		}
		else
		{
			error = _invalid = true;
			_failure_reason = "device name/attribute name \"" + dev_and_att + "\""
				+ "not in the format "
				"  \"domain/family/member/attrname\"";
		}
			
		
		/* if there are no errors, split the attribute list into the elements of
		 * the map. Until now there are attname and device name.
		*/
		if(!error)
		{
			elems = split(attlist, ',');
		
			for(i = 0; i < elems.size(); i++)
			{
// 				printf("\e[1;32m* \e[0mcreating attribute \"%s\" that maps\n"
// 					"  the attribute \"%s\" of the device \"%s\"\n",
//      					elems[i].c_str(), attname.c_str(), devname.c_str());
				attr = createAttribute(devname, attname, elems[i], i);
				attrs.push_back(attr);
			}
		}
		
	}
	
}

MappedAttr* MapAttributeAllocator::createAttribute(string devname, string attname, 
		string map, int index)
{
		DeviceProxy* dev;
		char failure[256];
		if(devname == "")
		{
			snprintf(failure, 256,
				 "Device name not specified for attribute \"%s\" "
				"  which should be mapped into \"%s\""
			        "  Try specifying the property \"TangoDevice\" in the device"
			        "  properties or be sure to have provided a tango device name"
			        " for each attribute to be mapped"
			        "  \"Map\" class property ",attname.c_str(),devname.c_str());
			_invalid = true;
			_failure_reason = string(failure);
			return NULL;
		}
		if(attname == "")
		{
// 			printf("\e[1;4;36minfo\e[0m: skipping element %d of attribute \"%s\"\n"
// 					"  of the device \"%s\"\n", index, attname.c_str(), 
// 					devname.c_str());
			return NULL; /* but still valid */
		}
		try
		{
			dev = new DeviceProxy(devname);
			try
			{
				AttributeInfoEx ai = dev->attribute_query(attname);
				AttrWriteType wt = ai.writable;
				AttrDataFormat df = ai.data_format;
				int data_t = ai.data_type;
				int max_x = ai.max_dim_x;
				
				if(max_x < index)
				{
					snprintf(failure, 256, "the index %d is greater than the"
						"  attribute maximum x dimension (%d)", index,
      						max_x);
					_invalid = true;
					_failure_reason = string(failure);
					return NULL;
				}
				MappedAttr *ma = new MappedAttr(map.c_str(), data_t, wt, df, 
						devname, attname, index);
				_invalid = false;
				delete dev;
				return ma;
			}
			catch(DevFailed &e)
			{
				snprintf(failure, 256, "failed to obtain attribute information"
					"  about \"%s\" while mapping into \"%s\""
					"%s ",
     					attname.c_str(), map.c_str(), 
						MappedAttr::exception("", e).c_str());
				_invalid = true;
				_failure_reason = string(failure);
				return NULL;
			}
		}
		catch(DevFailed &e)
		{
			snprintf(failure, 256, "failed to open device \"%s\""
				"  while mapping \"%s\" into \"%s\"%s", devname.c_str(), 
					attname.c_str(), map.c_str(), 
			      			MappedAttr::exception("tango exception", e).c_str());
			_invalid = true;
			_failure_reason = string(failure);
			return NULL;
		}
		
}

vector<string> MapAttributeAllocator::split(string s, const char delim)
{
	int *delimiters;
	vector<string> ret;
	char el[ELEMLEN];
	unsigned int i, delimcnt = 0;
	for(i = 0; i < s.length(); i++)
	{
		if(s[i] == delim)
			delimcnt++;
	}
	delimiters = new int[delimcnt+2];
	delimiters[0] = -1;
	delimcnt = 1;
	
	for(i = 0; i < s.length(); i++)
	{
		if(s[i] == delim)
		{
			delimiters[delimcnt] = i;
			delimcnt++;
		}
	}	
	delimiters[delimcnt] = s.length();
	
	for(i = 0; i < delimcnt; i++)
	{
		memset(el, 0, ELEMLEN);
		if(delimiters[i+1] - delimiters[i] - 1 < ELEMLEN)
		{
			strncpy(el, s.c_str() + delimiters[i] + 1, delimiters[i+1] - delimiters[i] - 1);
			ret.push_back(el);
		}
		else
		{
// 			printf("\e[1;31m*\e[0m the element %d (between positions %d and"
// 					" %d) is too long (%d characters)!\n"
// 					   "  The maximum size for an attribute name is %d\n",
// 				ret.size() + 1,  delimiters[i] + 1, delimiters[i+1],
// 					delimiters[i+1] - delimiters[i] - 1,  ELEMLEN);
			exit(EXIT_FAILURE);
		}
	}
		
	delete delimiters;
	return ret;
}



} /* namespace */




