static const char *RcsId = "$Header: /home/cvsadm/cvsroot/elettra/server/tangodevmapper/src/TangoDevMapper.cpp,v 1.5 2017-07-31 13:01:10 claudio Exp $";
//+=============================================================================
//
// file :         TangoDevMapper.cpp
//
// description :  C++ source for the TangoDevMapper and its commands. 
//                The class is derived from Device. It represents the
//                CORBA servant object which will be accessed from the
//                network. All commands which can be executed on the
//                TangoDevMapper are implemented in this file.
//
// project :      TANGO Device Server
//
// $Author: claudio $
//
// $Revision: 1.5 $
//
// $Log: TangoDevMapper.cpp,v $
// Revision 1.5  2017-07-31 13:01:10  claudio
// ticket #1554: porting to tango-9.2.5.a/QTANGO_BASE_CLASS
//
// Revision 1.4  2013-06-27 15:17:41  claudio
// pulizia alcuni warning
//
// Revision 1.3  2013-04-03 13:34:25  claudio
// ticket 729: go to ON state after successufl init
// use tango log instead of printf
//
// Revision 1.2  2012-05-07 10:33:52  claudio
// tango-7.2.6
// meno printf
//
// Revision 1.1.1.1  2008/04/28 09:17:53  giacomo
// Prima release di TangoDevMapper. Aprile 2008
//
//
// copyleft :     European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                FRANCE
//
//-=============================================================================
//
//  		This file is generated by POGO
//	(Program Obviously used to Generate tango Object)
//
//         (c) - Software Engineering Group - ESRF
//=============================================================================



//===================================================================
//
//	The following table gives the correspondence
//	between commands and method name.
//
//  Command name|  Method name
//	----------------------------------------
//  State   |  dev_state()
//  Status  |  dev_status()
//
//===================================================================


#include <tango.h>
#include <TangoDevMapper.h>
#include <TangoDevMapperClass.h>
#include "MappedAttr.h"
#include "MapAttributeAllocator.h"

using namespace std;	
			 
namespace TangoDevMapper_ns
{


	
//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::TangoDevMapper(string &s)
// 
// description : 	constructor for simulated TangoDevMapper
//
// in : - cl : Pointer to the DeviceClass object
//      - s : Device name 
//
//-----------------------------------------------------------------------------
TangoDevMapper::TangoDevMapper(Tango::DeviceClass *cl,string &s)
:TANGO_BASE_CLASS(cl,s.c_str())
{
	init_device();
}

TangoDevMapper::TangoDevMapper(Tango::DeviceClass *cl,const char *s)
:TANGO_BASE_CLASS(cl,s)
{
	init_device();
}

TangoDevMapper::TangoDevMapper(Tango::DeviceClass *cl,const char *s,const char *d)
:TANGO_BASE_CLASS(cl,s,d)
{
	init_device();
}
//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::delete_device()
// 
// description : 	will be called at device destruction or at init command.
//
//-----------------------------------------------------------------------------
void TangoDevMapper::delete_device()
{
	//	Delete device's allocated object
	set_state(Tango::CLOSE);
	int cnt = 0;
	set_status("deallocating attributes");
	std::map<string, MappedAttr *>::iterator iter;
	Database *db = new Database();
	DbData db_data; 
	DbDatum header("     -----     ");
	DbDatum attributeName("attributeName"), assocDevName("associatedDevice"),
			assocAttributeName("assocAttName"), associatetIsSpectrum("assocAttrIsSpectrum"),
			assocSpectrumIndex("assocSpectrumIndex");

	for(iter = attributesMap.begin(); iter != attributesMap.end(); iter++)
	{
		cnt++;
		MappedAttr *ma = iter->second;
		if(ma != NULL)
		{
			DbDatum attName(iter->first.c_str());
			db_data.push_back(attName);
			db_data.push_back(header);
			db_data.push_back(attributeName);
			db_data.push_back(assocDevName);
			db_data.push_back(assocAttributeName);
			db_data.push_back(associatetIsSpectrum);
			db_data.push_back(assocSpectrumIndex);
// 			printf("\r                                                                                       \r");
// 			printf("- \"%s\"\t", ma->name().c_str());
			fflush(stdout);
// 			printf("[\e[1;32mok\e[0m]");
			fflush(stdout);
			db->delete_device_attribute_property(this->get_name().c_str(), db_data);
			db_data.clear();
		}
	}
// 	printf("\r                                                                                       \r");
// 	printf("  \e[0m done deleting descriptive attribute properties");
// 	printf(" (%d entries.)\n", cnt);
	std::map<string, DeviceProxy*>::iterator iter2;
	for(iter2 =  namesDevsMap.begin();  iter2 != namesDevsMap.end(); iter2++)
	{
		if(iter2->second != NULL)
			delete iter2->second;
	}
	namesDevsMap.clear();
	attributesMap.clear();
	delete db;
}

//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::init_device()
// 
// description : 	will be called at device initialization.
//
//-----------------------------------------------------------------------------
void TangoDevMapper::init_device()
{
	INFO_STREAM << "TangoDevMapper::TangoDevMapper() create device " << device_name << endl;

	// Initialise variables to default values
	//--------------------------------------------
	unsigned int i, j, cnt = 0, mappednum = 0;
	get_device_property();
	string device;
	for(i = 0; i < map.size(); i++)
	{
		MapAttributeAllocator allocator(map[i], tangoDevice);
		if(allocator.isValid())
		{
			cnt++;
			vector<MappedAttr *> mapped = allocator.attributes();
			for(j = 0; j < mapped.size(); j++)
			{
				if(mapped[j] != NULL)
				{
					if(addAttributeToAttributeMap(mapped[j]))
					{
						add_attribute(mapped[j]);
						setupDevice(mapped[j]);
						mappednum++;
					}
					else
					{
						FATAL_STREAM<< "duplicate attribute name "<<mapped[j]->attributeName()<<endl;
						exit(EXIT_FAILURE);
					}
				}
			}
		}
		else
		{
 			ERROR_STREAM<<"error processing entry: "<<map[i]<<" "<<allocator.failureReason()<<endl;
		}
	}
	set_state(Tango::ON);

}

void TangoDevMapper::setupAttributesProperties()
{
	//int i;
	Tango::Database *db = new Tango::Database();
	long int index;
	string deviceName = get_name();
	char *prop1, *prop2, *prop3, *prop4;
	std::map<string, MappedAttr *>::iterator iter;
	//char description[512];
	
	for(iter = attributesMap.begin(); iter  != attributesMap.end(); iter++)
	{
		MappedAttr *a = iter->second;
		if(a == NULL)
		{
			INFO_STREAM<<"skipping a NULL attribute"<<endl;
			continue;
		}
		//char ind[32];
		DbData db_data; 
		
		DbDatum attName(a->name());
		DbDatum header("     -----     ");
		DbDatum attributeName("attributeName"), assocDevName("associatedDevice"),
			assocAttributeName("assocAttName"), associatetIsSpectrum("assocAttrIsSpectrum"),
			assocSpectrumIndex("assocSpectrumIndex");
		prop1 = new char[a->name().length() + 1];
		strncpy(prop1, a->name().c_str(), a->name().length() + 1);
		attributeName << a->name().c_str();
		prop2 = new char[a->deviceName().length() + 1];
		strncpy(prop2, a->deviceName().c_str(), a->deviceName().length() + 1);
		assocDevName << prop2;
		prop3 = new char[a->attributeName().length() + 1];
		strncpy(prop3, a->attributeName().c_str(), a->attributeName().length() + 1);
		assocAttributeName << prop3;
		
		if(a->dataFormat() == Tango::SPECTRUM)
		{
			attName << (unsigned short) 6;	
			header << DevString("The following 5 properties are descriptive. Changes will be ignored and lost at next startup");
			db_data.push_back(attName);
			db_data.push_back(header);
			db_data.push_back(attributeName);
			db_data.push_back(assocDevName);
			db_data.push_back(assocAttributeName);
			index = a->index();
			assocSpectrumIndex << index;	
			prop4 = new char[4];
			strncpy(prop4, "yes", 4);
			associatetIsSpectrum<<  prop4;
			db_data.push_back(associatetIsSpectrum);
			db_data.push_back(assocSpectrumIndex);
		}
		else
		{
			attName << (unsigned short) 5;
			header << DevString("The following 4 properties are descriptive. Changes will be ignored and lost at next startup");
			db_data.push_back(attName);
			db_data.push_back(header);
			db_data.push_back(attributeName);
			db_data.push_back(assocDevName);
			db_data.push_back(assocAttributeName);
			char isScalar[16] = "no, is scalar";
			prop4 = new char[strlen(isScalar) + 1];
			strncpy(prop4, isScalar, strlen(isScalar) + 1);
			associatetIsSpectrum<<  prop4;
			db_data.push_back(associatetIsSpectrum);
			
		}
		
		try{
// 			db->set_timeout_millis(10000);
			db->put_device_attribute_property(deviceName.c_str(), db_data);
		}
		catch(DevFailed &e)
		{
			ERROR_STREAM<<"failed inserting attribute properties into" << deviceName <<" for  "<< a->attributeName()<<endl;
		}
		
	}
}


//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::get_device_property()
// 
// description : 	Read the device properties from database.
//
//-----------------------------------------------------------------------------
void TangoDevMapper::get_device_property()
{
	//	Initialize your default values here (if not done with  POGO).
	//------------------------------------------------------------------

	//	Read device properties from database.(Automatic code generation)
	//------------------------------------------------------------------
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("Map"));
	dev_prop.push_back(Tango::DbDatum("TangoDevice"));

	//	Call database and extract values
	//--------------------------------------------
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_property(dev_prop);
	Tango::DbDatum	def_prop, cl_prop;
	TangoDevMapperClass	*ds_class =
		(static_cast<TangoDevMapperClass *>(get_device_class()));
	int	i = -1;

	//	Try to initialize Map from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  map;
	else {
		//	Try to initialize Map from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  map;
	}
	//	And try to extract Map value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  map;

	//	Try to initialize TangoDevice from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  tangoDevice;
	else {
		//	Try to initialize TangoDevice from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  tangoDevice;
	}
	//	And try to extract TangoDevice value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  tangoDevice;



	//	End of Automatic code generation
	//------------------------------------------------------------------
	if(tangoDevice == ""){
// 		printf("\e[1;33m*\e[0m warning: no generic device specified in the\n"
// 				"  device property \"TangoDevice\": it is supposed that you\n"
// 				"  provide a tango device for each \"Map\" element in the\n"
// 				"  device properties, otherwise the server \e[1;4;37mwill not "
// 				"start\e[0m\n");
	}
}
//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::always_executed_hook()
// 
// description : 	method always executed before any command is executed
//
//-----------------------------------------------------------------------------
void TangoDevMapper::always_executed_hook()
{
	
}
//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::read_attr_hardware
// 
// description : 	Hardware acquisition for attributes.
//
//-----------------------------------------------------------------------------
void TangoDevMapper::read_attr_hardware(vector<long> &attr_list)
{
	DEBUG_STREAM << "TangoDevMapper::read_attr_hardware(vector<long> &attr_list) entering... "<< endl;
	//	Add your own code here
}


bool TangoDevMapper::is_MappedAttr_allowed(Tango::AttReqType type)
{
	return true;
	
}

void TangoDevMapper::setupDevice(MappedAttr * att)
{
	string devname = att->deviceName();
	DeviceProxy* dev = findOpenDevices(devname);
	if(dev == NULL)
	{
		try
		{
			dev = new DeviceProxy(devname);
			/* add it to the map */
			namesDevsMap[devname] = dev;
		}
		catch(DevFailed &e)
		{
			ERROR_STREAM<<"failed DevProxy for "<<devname<<e<<endl;
		}
	}
	else{
 		DEBUG_STREAM<< "device " <<devname <<" already open"<<endl;
	}
}

DeviceProxy *TangoDevMapper::findOpenDevices(string dname)
{
	std::map<string, DeviceProxy *>::iterator iter = namesDevsMap.find(dname);
	if(iter != namesDevsMap.end())
	{
		return iter->second;
	}
	return NULL;
}

bool TangoDevMapper::addAttributeToAttributeMap(MappedAttr *att)
{
	std::map<string, MappedAttr*>::iterator iter = attributesMap.find(att->
			name());
	if(iter != attributesMap.end())
	{
		/* found an attribute with that name: bad */
		return false;
	}
	else
	{
		attributesMap[att->name()] = att;
		return true;
	}
}



}	//	namespace
