#include "TangoDevMapper.h"
#include "MappedAttr.h"

using namespace std;
using namespace Tango;

			 
namespace TangoDevMapper_ns
{

//+----------------------------------------------------------------------------
//
// method : 		TangoDevMapper::read_MappedAttr
// 
// description : 	Extract real attribute values for MappedAttr acquisition result.
//
//-----------------------------------------------------------------------------
void TangoDevMapper::read_MappedAttr(Tango::Attribute &attr)
{
	DEBUG_STREAM << "TangoDevMapper::read_MappedAttr(Tango::Attribute &attr) entering... "<< endl;
	
	char errmsg[512];
	bool wrong_dtype = false;
	long int dt; /* data type */
	int offset; /* used in SPECTRUM extraction, READ_WRITE */
	int size = 1;
	int index;
	Tango::AttrWriteType wt; /* write type (READ or READ_WRITE) */
	/* for extraction */
	DevUShort ush1;
	DevUChar uch1, uchar1;
	DevDouble d1;
	DevLong l1;
	DevFloat f1;
	DevShort sho;
	DevBoolean b1;
	
	/* to extract the READ_WRITE values or the SPECTRUMs */
	vector<DevDouble> d;
	vector<DevFloat> f;
	vector<DevLong> l;
	vector<DevULong> ul;
	vector<Tango::DevBoolean> b;
	vector<DevUChar> uc;
	vector<DevUShort> ush;
	vector<DevShort> sh;
	
	/* the device attribute to read */
	DeviceAttribute da;
	/* the pointer to the device proxy */
	DeviceProxy *dev;
	/* name of the attribute the user wants to read (attr) */
	string mappedname = attr.get_name();	
	std::map<string, MappedAttr*>::iterator iter = attributesMap.find(mappedname);
	if(iter == attributesMap.end())
	{
		Except::throw_exception("Error reading scalar attribute " + mappedname,
				"Did not find the device attribute associated to the mapped attribute " +mappedname,
				"TangoDevMapper::read_MappedAttr");
		return;
	}
	
	MappedAttr* mattr = iter->second;
	/* found an attribute with that name: good */
	/* now find the Tango device associated */
	string dname = mattr->deviceName();
	string attname = mattr->attributeName();
	std::map<string, DeviceProxy *>::iterator iter2 = namesDevsMap.find(dname);
	if(iter2 == namesDevsMap.end())
	{
		Except::throw_exception("Error reading scalar attribute " + attr.get_name(),
				"Did not find the device associated to the attribute " + attname,
				"TangoDevMapper::read_MappedAttr");
		return;
	}

	dev =  iter2->second;
	/* read_attribute() on the associated attribute with name 
	 * mattr->attributeName() */

	if(dev == NULL)
	{
		Except::throw_exception("Error reading scalar attribute " + attr.get_name(),
				"The device pointer is NULL",
				"TangoDevMapper::read_MappedAttr");
		return;
	}
	try
	{
		da = dev->read_attribute(attname);
	}
	catch(DevFailed &e)
	{
// 		printf("\r                                                                       \r");
// 		printf("\e[1;31m* \e[0mfailed reading %s from %s \e[1;34m-->\e[0m%s\n",
// 				attname.c_str(), dname.c_str(), attr.get_name().c_str());
		Except::re_throw_exception(e, "Error reading scalar attribute " + attr.get_name(),
				"read_attribute on device " + dname + " failed",
				"TangoDevMapper::read_MappedAttr");
	}
	/* get the pointer to the MultiAttribute list */
	MultiAttribute *ma;
	try
	{
		ma = get_device_attr();
	}
	catch(DevFailed &e)
	{
// 		printf("\r                                                                       \r");
// 		printf("\e[1;31m* \e[0mfailed getting MultiAttribute list %s\n",
// 				dname.c_str() );
		Except::re_throw_exception(e, "Error reading MultiAttribute list from " +dname,
				"get_device_attr() on device " + dname + " failed",
				"TangoDevMapper::read_MappedAttr");
		return;
	}
	
	/* store data type */
	dt = mattr->dataType();
	/* store write type */
	wt = mattr->writeType();
	
	/* must recognize the data type to extract it correctly */
	if(mattr->dataFormat() == Tango::SCALAR)
	{
		if(dt == Tango::DEV_DOUBLE)
		{
			if(wt == Tango::READ_WRITE)
			{
				da >> d;
				if(d.size() == 2)
				{
					try
					{
						DevDouble *dd = new DevDouble;
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(d[1]);
						*dd = d[0];
						attr.set_value(dd, 1, 0, true);
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
								"get_w_attr_by_name() on device " + dname + " failed",
								"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;
			}
			else if(wt == Tango::READ)
			{
				
				DevDouble *dd = new DevDouble;
				da >> d1;
				*dd = d1;
				attr.set_value(dd, 1, 0, true);
			}
			else
				wrong_dtype = true;
		}
		else if(dt == Tango::DEV_FLOAT)
		{	
			if(wt == Tango::READ_WRITE)
			{
				da >> f;
				if(f.size() == 2)
				{
					try
					{
						DevFloat *df = new DevFloat;
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(f[1]);
						*df = f[0];
						attr.set_value(df, 1, 0, true);
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
							"get_w_attr_by_name() on device " + dname + " failed",
							"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;
			}
			else if(wt == Tango::READ)
			{
				DevFloat *df = new DevFloat;
				da >> f1;
				*df = f1;
				attr.set_value(df, 1, 0, true);
			}
			else
				wrong_dtype = true;
		}	

		/* LONG */
		else if(dt == Tango::DEV_LONG)
		{
			DevLong *dl = new DevLong;
			if(wt == Tango::READ_WRITE)
			{
				da >> l;
				if(l.size() == 2)
				{
					try
					{
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(l[1]);
						*dl = l[0];
						attr.set_value(dl);
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
								"get_w_attr_by_name() on device " + dname + " failed",
								"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;
			}
			else if(wt == Tango::READ)
			{
				da >> l1;
				*dl = l1;
				attr.set_value(dl, 1, 0, true);
			}
			else
				wrong_dtype = true;
		}	

		else if(dt == Tango::DEV_STRING)
		{
			// 						da >> outstr;
			// 						str = new DevString;
			// 						*str = CORBA::string_dup(outstr.c_str());
			// 						attr.set_value(str, 1, 0, true);
		}

		else if(dt == Tango::DEV_BOOLEAN)
		{	
			Tango::DevBoolean *db = new Tango::DevBoolean;
			if(wt == Tango::READ_WRITE)
			{
				da >> b;
				if(b.size() == 2)
				{
					try
					{
                        Tango::DevShort bo_local;
                        bo_local = b[1];
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(bo_local);
                            
						*db = b[0];
						attr.set_value(db);
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
								"get_w_attr_by_name() on device " + dname + " failed",
								"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;
			}
			else if(wt == Tango::READ)
			{
				da >> b1;
				*db = b1;
// 				printf("tango READ: boolean\n");
				attr.set_value(db);
			}
			else
				wrong_dtype = true;
		}	
		else if(dt == Tango::DEV_UCHAR)
		{
			DevUChar *duc = new DevUChar;
			if(wt == 
					Tango::READ_WRITE)
			{
				da >> uc;
				if(uc.size() == 2)
				{
					try
					{
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(uc[1]);
						attr.set_value(&( uc[0]));
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
								"get_w_attr_by_name() on device " + dname + " failed",
								"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;
			}
			else if(wt == Tango::READ)
			{
				da >> uchar1;
				*duc = uchar1;
// 				printf("tango READ: %f\n",  uchar1);
				attr.set_value(duc, 1, 0, true);
			}
			else if(wt == Tango::WRITE)
			{
				da >> uch1;
				*duc = uch1;
// 				printf("tango write: %f\n",  uch1);
			}
			else
				wrong_dtype = true;
		}

		/* SHORT */	
		else if(dt == Tango::DEV_SHORT)
		{
			if(wt == 
					Tango::READ_WRITE)
			{
				da >> sh;
				if(sh.size() == 2)
				{
					try
					{
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(sh[1]);
						attr.set_value(&( sh[0]));
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
								"get_w_attr_by_name() on device " + dname + " failed",
								"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;

			}
			else if(wt == Tango::READ)
			{
				DevShort *sh1 = new DevShort;
				da >> sho;
				*sh1 = sho;
// 				printf("\n++tango short scalar READ: %d\n",  sho);
				attr.set_value(sh1, 1, 0, true);
				return;
			}
			else if(wt == Tango::WRITE)
			{
				// 							da >> sho;
				// 							printf("tango write: %f\n",  sh1);
			}
			else
				wrong_dtype = true;
		}

		else if(dt == Tango::DEV_USHORT)
		{
			if(wt == 
					Tango::READ_WRITE)
			{
				da >> ush;
				if(ush.size() == 2)
				{
					try
					{
						ma->get_w_attr_by_name(attr.get_name().c_str()).
							set_write_value(ush[1]);
						attr.set_value(&( ush[0]));
					}
					catch(DevFailed &e)
					{
						Except::re_throw_exception(e, "Error getting WAttribute list for " +dname,
								"get_w_attr_by_name() on device " + dname + " failed",
								"TangoDevMapper::read_MappedAttr");
					}
				}
				else
					wrong_dtype = true;
			}
			else if(wt == Tango::READ)
			{
				DevUShort *dush = new DevUShort;
				da >> ush1;
				*dush = ush1;
				//printf("tango READ: %f\n",  ush1);
				attr.set_value(dush, 1, 0, true);
			}
			else if(wt == Tango::WRITE)
			{
				da >> ush1;
				DEBUG_STREAM<<"tango write: "<< ush1<<endl;
			}
			else
				wrong_dtype = true;
		}	
		else
		{
			ERROR_STREAM<<"reading "<< attname.c_str()<<" from "<<dname.c_str() <<"-->"<< attr.get_name().c_str()<<endl;			
			Except::throw_exception("Error reading scalar attribute " + attr.get_name(),
					"read_attribute on device " + dname + " failed: unsupported data Type",
					"TangoDevMapper::read_MappedAttr");	
		}

	}
	else if(mattr->dataFormat() == Tango::SPECTRUM)
	{
		index = mattr->index();
		bool error = false;
		
		if(dt == Tango::DEV_DOUBLE)
		{
				da >> d;
				if(wt == Tango::READ_WRITE && (size = d.size()/2) > index )
				{
					DevDouble *dd = new DevDouble;
					offset = d.size() / 2;
					*dd = d[index];
					attr.set_value(dd, 1, 0, true);
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(d[offset + index]);
				}
				else if(wt == Tango::READ && (size = d.size()) > index)
				{
					DevDouble *dd = new DevDouble;
					*dd = d[index];
					attr.set_value(dd, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}
		else if(dt == Tango::DEV_FLOAT)
		{
				da >> f;
				if(wt == Tango::READ_WRITE && (size = f.size()/2)  > index )
				{
					DevFloat *ff = new DevFloat;
					offset = f.size() / 2;
					*ff = f[index];
					attr.set_value(ff, 1, 0, true);
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(f[offset + index]);
				}
				else if(wt == Tango::READ && (size = f.size())  > index)
				{
					DevFloat *ff = new DevFloat;
					*ff = f[index];
					attr.set_value(ff, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}
		else if(dt == Tango::DEV_LONG)
		{
				da >> l;
				if(wt == Tango::READ_WRITE &&  (size = l.size()/2)  > index )
				{
					DevLong *ll = new DevLong;
					offset = l.size() / 2;
					*ll = l[index];
					attr.set_value(ll, 1, 0, true);
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(l[offset + index]);
				}
				else if(wt == Tango::READ && (size = l.size())  > index)
				{
					DevLong *ll = new DevLong;
					*ll = l[index];
					attr.set_value(ll, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}
		/* boolean */
		else if(dt == Tango::DEV_BOOLEAN)
		{
				da >> b;
				if(wt == Tango::READ_WRITE &&  (size = b.size()/2)  > index )
				{
					DevBoolean *db = new DevBoolean;
					offset = b.size() / 2;
					*db = b[index];
					attr.set_value(db, 1, 0, true);
                    Tango::DevShort bval = b[offset + index];
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(bval);
				}
				else if(wt == Tango::READ && (size = b.size())  > index)
				{
					DevBoolean *db = new DevBoolean;
					*db= b[index];
					attr.set_value(db, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}
		else if(dt == Tango::DEV_UCHAR)
		{
				da >> uc;
				if(wt == Tango::READ_WRITE &&  (size = uc.size()/2) > index )
				{
					DevUChar *duc = new DevUChar;
					offset = uc.size() / 2;
					*duc = uc[index];
					attr.set_value(duc, 1, 0, true);
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(uc[offset + index]);
				}
				else if(wt == Tango::READ && (size = uc.size())  > index)
				{
					DevUChar *duc = new DevUChar;
					*duc= uc[index];
					attr.set_value(duc, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}
		else if(dt == Tango::DEV_SHORT)
		{
				da >> sh;
				if(wt == Tango::READ_WRITE &&  (size = sh.size()/2)  > index )
				{
					DevShort *dsh = new DevShort;
					offset = sh.size() / 2;
					*dsh = sh[index];
					attr.set_value(dsh, 1, 0, true);
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(sh[offset + index]);
				}
				else if(wt == Tango::READ &&(size = sh.size())  > index)
				{
					DevShort *dsh = new DevShort;
					*dsh= sh[index];
					attr.set_value(dsh, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}
		else if(dt == Tango::DEV_USHORT)
		{
				da >> ush;
				if(wt == Tango::READ_WRITE &&  (size = ush.size()/2)  > index )
				{
					DevUShort *dush = new DevUShort;
					offset = ush.size() / 2;
					*dush = ush[index];
					attr.set_value(dush, 1, 0, true);
					ma->get_w_attr_by_name(mappedname.c_str())
							.set_write_value(ush[offset + index]);
				}
				else if(wt == Tango::READ &&  (size = ush.size())  > index)
				{
					DevUShort *dush = new DevUShort;
					*dush= ush[index];
					attr.set_value(dush, 1, 0, true);
				}
				else
					wrong_dtype = true;
		}		
		else
		{
				ERROR_STREAM<<"reading "<< attname.c_str()<<" from "<<dname.c_str() <<"-->"<< attr.get_name().c_str()<<endl;	
				Except::throw_exception("Error reading scalar attribute " + attr.get_name(),
						"read_attribute on device " + dname + " failed: unsupported data Type",
						"TangoDevMapper::read_MappedAttr");

		}
		if(error)
		{
// 			snprintf(errmsg, 512, "\e[1;31m* \e[0mfailed reading %s from %s \e[1;34m-->\e[0m%s\n"
// 					"  \e[1;4;31merror\e[0m: mapped index [%d] exceeds target attribute dimensions [%d]\n"
// 					"  \e[1;4;35mhint\e[0m: check the target attribute \"%s\" dimensions\n"
// 					"  on the device \"%s\" and revise the map for the attribute\n"
// 					"  \"%s\" in this device \e[1;4;37mdevice properties\e[0m so that the\n"
// 					"  list of attributes does not exceed the size %d\n",
// 					attname.c_str(), dname.c_str(), attr.get_name().c_str(),
// 					index, size, attname.c_str(), dname.c_str(),attr.get_name().c_str() );
// 			printf(errmsg);
			Except::throw_exception("Error reading scalar attribute " + attr.get_name(),
					string(errmsg),
					"TangoDevMapper::read_MappedAttr");		
		}
		else
		{
// 			printf("\r                                                                       \r");
// 			printf("\e[1;32m* \e[0mread %s[%d/%d] from %s \e[1;34m--> \e[0m%s",
// 					attname.c_str(), index, size, dname.c_str(), attr.get_name().c_str());
// 			fflush(stdout);

		}
	}
	else
	{	
		Except::throw_exception("Error decoding scalar attribute " + attr.get_name(),
				"The attribute data format is not supported: supported "
				"formats are SCALAR and SPECTRUM",
				"TangoDevMapper::read_MappedAttr");
		return;
	}

	if(wrong_dtype)
	{
// 		snprintf(errmsg, 512, "\e[1;31m* \e[0mfailed reading %s from %s (index %d in size %d) \e[1;34m-->\e[0m%s\n"
// 				"  \e[1;4;31merror\e[0m: wrong data type or error in extracting\n"
// 				"  the read/set points.\n"
// 				"  If the attribute is a spectrum, a mismatch in the extracted data sizes could have\n"
// 				"  happened, although this is strange. Contact the author\n",
// 				attname.c_str(), dname.c_str(), index, size, attr.get_name().c_str());
// 		printf(errmsg);
		Except::throw_exception("Error reading scalar attribute - wrong_type " + attr.get_name(),
				string(errmsg),
				"TangoDevMapper::read_MappedAttr");
	}

}

}













